/*
**	WSWAP - Select lines to include/exclude from file
**
**	Try compiling with Microsoft C versions 5.1 or above.
**		qcl /DMSDOS wswap.c /c
**		link /stack:10000 wswap+\msc51\lib\binmode /NOE;
**	DISCLAIMER: The code may not be elegant, but it does the job.
**
**
**	Modification History
**	1.0	01-Aug-94	Steve Karra
**	1.1	01-Jan-95	Check for Disk Full condition
**				Allow for wildcards and case insensitivity
**	1.2	16-Feb-95	OPTION keyword implemented
*/

#include <stdio.h>
#include <ctype.h>
#include <conio.h>
#include <io.h>
#include <string.h>
#include <stdlib.h>

#define xstrcmp(a,b) xstrncmp(a,b,128)

#define  TSWAP        "SWAP"
#define  TOPT         "OPTION"
#define  TEND         "END"
#define  TDEFAULT     "DEFAULT"
#define  VAR          '%'
#define  PLUS         '+'
#define  MINUS        '-'
#define  STAR         '*'
#define  ARROW        '^'
#define  BANG         '!'
#define  DEBUG        '='
#define  BELL         '\007'
#define  CMDLEN       256
#define  MAXLINE      256
#define  MAXFN        128
#define  NTOKENS      64	/* 0 thru NTOKENS-3 can be used */
#define  TOKLEN       32
#define  TTYPE        token[NTOKENS-1]
#define  TSELECT      token[NTOKENS-2]


static char *Version="WSWAP Version 1.2 (written by Steve Karra)";
FILE  *ifptr, *ofptr;
char  token[NTOKENS][TOKLEN];
char  swapstr[TOKLEN], optstr[TOKLEN];
char  remarkstr[TOKLEN], remarkstr2[TOKLEN];
int   pflag=0, mflag=0, sflag=0, cflag=0, dflag=0, bflag=0;
char  *fn;


abortpgm(n,str)
int  n;
char *str;
{
	printf("File: %s\n", fn);
	switch(n) {
	    case(1):	printf("EOF Reached with no END statement\n");
			break;
	    case(2):	printf("Line: %s", str);
			printf("Maximum of %d tokens exceeded\n",(NTOKENS-2));
			break;
	    case(3):	printf("Line: %s", str);
			printf("END missing to previous selection\n");
			break;
	    case(4):	printf("Cannot delete original file [%s]\n", str);
			break;
	    case(5):	printf("Cannot rename temp file [%s] to original deleted file\n", str);
			break;
	    case(6):	printf("Cannot copy temp file [%s] to original\n", str);
			break;
	    case(7):	printf("Length of output file [%s] varies too much - no changes effected\n", str);
			break;
	    case(8):	printf("Cannot write line to output file [%s]\n", str);
			break;
	    case(9):	printf("EOF Reached with no extra OPTION line\n");
			break;
	    case(10):	printf("OPTION Variable error (no terminating %c): %s\n", VAR, str);
			break;
	    default:	break;
	    }
	printf("Hit any key to continue ...");
	do {} while (!kbhit());
	while (kbhit()) getch();
	exit(-1);
}


gettokens(line, sstr)
char *line, *sstr;
{
    int  i;
    char *cstr, *ch;

    for (i=0; i<(NTOKENS-2); i++) token[i][0]='\0';
    ch=&line[strlen(sstr)];  i=0;

    do {
    	cstr=&token[i][0];
    	while (*ch==' ') ch++;
    	while ((*ch!=' ')&&(*ch!='\0')&&(*ch!='\n')&&(*ch!='\r')) *(cstr++)=*(ch++);
	*cstr='\0';  i++;
	if (i>=(NTOKENS-2)) abortpgm(2);
    } while ((*ch!='\0')&&(*ch!='\n')&&(*ch!='\r'));
}


int xstrncmp(a,b,l)
char *a, *b;
int l;
{
    int i=1;

    while ((l>i) && ((*a==*b)||(cflag&&(toupper(*a)==toupper(*b)))) &&
	(*a!='\0') && (*b!='\0')) a++,b++,i++;
    if ((l==i)&&(*a==*b)) return(0);
    if ((*a=='*')||(*b=='*')||((*a=='\0')&&(*b=='\0'))) return(0);
    else return(i);
}


processopt(line,nf)
char *line, *nf;
{
    char nline[MAXLINE], *ch, *nch;
    int i;

    gettokens(line, optstr);
    if (dflag) printf("[%s] OPTION line found: comparing %s and %s\n", fn, token[0], TTYPE);
    if (xstrcmp(token[0],TTYPE)==0) {
	ch=&line[strlen(optstr)+strlen(TTYPE)+2];
	nch=nline; *nch='\0';
	while (*ch!='\0') {
	    if (*ch==VAR) {
		ch++; i=0; token[0][0]='\0';
		while ((*ch!=VAR)&&(*ch!='\0')) token[0][i++]=*ch++;
		if (*ch++!=VAR) abortpgm(10,token[0]);
		token[0][i]='\0';
		nch=getenv(token[0]);
		if (strlen(nch)==0) printf("[%s] \007Warning - environment variable %s could not be found\n", fn, token[0]);
		strcat(nline,nch);
		nch=&nline[strlen(nline)];
		}
	    else *nch++ = *ch++;
	    *nch = '\0';
	    }
	if (dflag) printf("[%s] Successful match: writing %s\n", fn, nline);
	if (fprintf(ofptr,"%s", nline)<=0) abortpgm(8,nf);
	if (fgets(line, MAXLINE, ifptr) == NULL) abortpgm(9,nf);
	}
}



main ( argc, argv )
int   argc;
char  *argv[];
{
    char  nf[MAXFN], line[MAXLINE], nline[MAXLINE], cmd[CMDLEN], *ch, *nch;
    int   onematch, match;
    int   remark, notfirst, option, endf, i, s1, s2, var, ifil;


    if (argc < 4) {
	fprintf(stderr,"%s\n", Version);
	fprintf(stderr,"%cUsage: %s TYPE SELECTION File_List\n", BELL, argv[0]);
	fprintf(stderr,"       File_List can optionally be prefixed with +,-,*,^,! or = (debug)\n");
	exit(-1);
	}


    strcpy(TTYPE,argv[1]);
    strcpy(TSELECT,argv[2]);

    for (ifil=3; ifil<argc; ifil++) {
	tmpnam(nf);	/* will be created on current drive */
	do {
	    if (*argv[ifil]==PLUS) pflag++,argv[ifil]++;
	    else if (*argv[ifil]==MINUS) mflag++,argv[ifil]++;
	    else if (*argv[ifil]==STAR) sflag++,argv[ifil]++;
	    else if (*argv[ifil]==ARROW) cflag++,argv[ifil]++;
	    else if (*argv[ifil]==DEBUG) dflag++,argv[ifil]++;
	    else if (*argv[ifil]==BANG) bflag++,argv[ifil]++;
	    else break;
	} while (1);
	fn=argv[ifil];

        if (((ifptr=fopen(argv[ifil],"r")) == NULL) ||
                ((ofptr=fopen(nf,"w")) == NULL)) {
            fprintf(stderr,"%c%s: Cannot open files %s and/or %s\n",BELL, 
		argv[0], argv[ifil], nf);
            continue;
	    }

	notfirst=0;
        while (fgets(line, MAXLINE, ifptr) != NULL) {

	    if (!notfirst) {
		strcpy(remarkstr, line);
		ch = remarkstr;
		while ((*ch!='\0')&&(*ch!='\n')&&(*ch!='\r')) ch++;
		*ch = '\0';
		strcpy(remarkstr2, remarkstr);
		strcat(remarkstr," ");
		strcat(remarkstr2,"\011");
		strcpy(swapstr, remarkstr);
		strcat(swapstr, TSWAP);
		strcpy(optstr, remarkstr);
		strcat(optstr, TOPT);
		notfirst++;
		}

	    /* Reading ordinary lines - just copy to output if not SWAP */
            if (xstrncmp(line,swapstr,strlen(swapstr))==0) {

		/* Now we have to check if its a valid TYPE - if not just copy to output */
                gettokens(line, swapstr);
                if (xstrcmp(token[0],TTYPE)==0) {

		    /* Its a match - now we loop until END.  Will check SELECT etc */
		    if (dflag) printf("[%s] Matched: %s\n", fn, TTYPE);
		    match=0;  onematch=0;

                    do {
		        /* If its the right type, uncomment, else comment */
			if (xstrncmp(line,swapstr,strlen(swapstr))==0) { /* special line */
			    if (dflag) printf("[%s] Matched Option: %s", fn, line);
			    gettokens(line, swapstr);
			    if (!sflag)
				if (fputs(line, ofptr)!=0) abortpgm(8,nf);
			    /* First check if a change */
                	    if (xstrcmp(token[0],TTYPE)==0) {
			        match=0;  endf=0;  i=0;
			        while ((!match)&&(i<(NTOKENS-2))&&(strlen(token[i])>0)) {
			    	    if (xstrcmp(token[i],TEND)==0) {
					endf++;
					break;
					}
			            if (xstrcmp(token[i],TSELECT)==0) match++,onematch++;
			            else if ((xstrcmp(token[i],TDEFAULT)==0)&&(!onematch)) {
					if (dflag) printf("[%s] Match fail warning - DEFAULT action taken\n", fn);
					match++;
					}
			            else match=0;
				    i++;
			            }
				if (endf) break;
			        }
			    else abortpgm(3,line);
			    }
			else {  /* not special - ordinary line */
			    /* check for REM<space> and REM<tab> */
			    remark = ((xstrncmp(line,remarkstr,strlen(remarkstr))==0)||
				(xstrncmp(line,remarkstr2,strlen(remarkstr2))==0));
			    option = (xstrncmp(line,optstr,strlen(optstr))==0);

			    if ((!match) && (mflag)) {  /* don't touch */
				ch = &line[0];
				}
			    else if ((match) && (!mflag)) { /* include these ones */
				ch = (remark&&(!option))?&line[strlen(remarkstr)]:&line[0];
				}
			    else {  /* don't include these */
				ch = &line[0];
				if (((!remark)&&(!pflag)&&(!sflag)) || (option&&(!sflag)))
				    if (fprintf(ofptr, "%s", remarkstr)<=0) abortpgm(8,nf);
			        }
			    /* time to print it out if required */
			    if ( (!match) && (sflag) ) ;
			    else {
				if ((!sflag)||(!option)) if (fputs(ch, ofptr)!=0) abortpgm(8,nf);
				if (xstrncmp(ch,optstr,strlen(optstr))==0) processopt(ch,nf);
				}
			    }

                        if (fgets(line, MAXLINE, ifptr)==NULL) abortpgm(1,"\0");
                        } while (!feof(ifptr));
		    if (feof(ifptr)) abortpgm(1,"\0");
                    }
                else if (fputs(line, ofptr)!=0) abortpgm(8,nf);  /* not a valid token so just copy */
                }

	    /* Look for OPTION key */
            else if (xstrncmp(line,optstr,strlen(optstr))==0) {
		if (!sflag) if (fputs(line, ofptr)!=0) abortpgm(8,nf);
		processopt(line,nf);
		}

	    else if (fputs(line, ofptr)!=0) abortpgm(8,nf);

            } /* end of while loop */

	/* Do an extra check */
	fclose(ofptr);	ofptr=fopen(nf,"r");
	s1=filelength(fileno(ifptr));	if (dflag) printf(" Input size: %d\n",s1);
	s2=filelength(fileno(ofptr));	if (dflag) printf("Output size: %d\n",s2);
	var=abs((int)(100*(s1-s2)/s1));
	if (dflag) printf("   Variance: %d%%\n",var);
	if ((var>5)&&(!bflag)) abortpgm(7,nf);

        fclose(ifptr);
        fclose(ofptr);

	/* We used to remove and rename, but now copy and remove */
	/* Changed because tmp files cannot be renamed across drives */
	strcpy(cmd,"copy ");  strcat(cmd,nf);  strcat(cmd," ");
	strcat(cmd,argv[ifil]);  strcat(cmd," /V > NUL:");
	if (system(cmd)) abortpgm(6,nf);
	if (remove(nf)!=0) abortpgm(4,nf);
/*
**	if (remove(argv[ifil])!=0) abortpgm(4,argv[ifil]);
**	if (rename(nf,argv[ifil])!=0) abortpgm(5,nf);
*/
    }

    exit(0);
}
